#!/usr/bin/env python3

from __future__ import print_function

"""
Class to read .gpx files

OpenCPN can export route data as .gpx file.
They provide only the waypoint locations, names and description.
We have to compute the distance and bearing from one
waypoint to the next.

We allow the output to be somewhat customizable.

GPX data does not unfortunately include the variation
at each waypoint, so we use cmweiss@gmail.com's geomag to calculate
variation. 
We provide it the latest WMM.COF data (WMM2015v2)
and it produces the same results as

https://www.ngdc.noaa.gov/geomag-web/calculators/calculateDeclination
"""

"""
RCSid:
	$Id: GPX.py,v 1.26 2023/07/11 23:53:32 sjg Exp $

	@(#) Copyright (c) 2019-2023, Simon J. Gerraty

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net

"""

import os
import os.path
import sys
import time
import xml.etree.ElementTree as ET
import logging
import latlong
import geomag
import deviation

# default fields
gpx_fields = 'Waypoint,Position,Course,Magnetic,Distance,cDistance,Remarks'

# default formats for fields
gpx_formats = {
    'compass': '  {:03.0f} C',
    'course': '{:03.0f} T',
    'deviation': '      {:02.1f}',
    'distance': '{:6.2f} NMi',
    'empty': '{:6}',
    'leg': '{:3d}',
    'magnetic': '   {:03.0f} M',
    'position': '{:20}',
    'speed': '{:2.1f} kts',
    'time': '{:5.1f} min',
    'variation': '     {:02.1f}',
    'waypoint': '{:15}',
    }

def gpx_format(field, conf={}):
    return conf.get('fmt:'+field, gpx_formats[field])

class GPX(object):
    """base class for GPX"""
    def __init__(self, parent=None, conf={}):
        if parent is None:
            self.fields = 'Name'
            self.conf = conf
            gpx_f = conf.get('gpx_f')
            self.node = self.root = ET.parse(gpx_f).getroot()
            self.creator = self.root.attrib['creator']
            self.name = conf.get('gpx_file')
            self.ns = self.root.tag.strip('gpx')
            for k in gpx_formats.keys():
                exec('GPX.fmt_{} = gpx_format("{}", self.conf)'.format(k,k))
            # we use this to compute variation
            GPX.gm = geomag.GeoMag()
            # for cLog (corrected Log)
            # and cDistance
            GPX.log_error = float(self.conf.get('log_error', 0))
            GPX.fmt = self.conf.get('fmt', 'verbose')
            if self.fmt == 'csv':
                sep = ','
                GPX.fmt_position = GPX.fmt_waypoint = '{}'
                GPX.hdr_distance = GPX.hdr_course = '{}'
            else:
                sep = ' '
            GPX.sep = self.conf.get('sep', sep)
            w = None
            for k in ['course','distance','speed','time']:
                exec('w = len(self.fmt_{}.format(1.0))'.format(k),globals(),locals())
                exec('GPX.hdr_{} = "{{:>{}}}"'.format(k, w),globals(),locals())
            d = self.conf.get('deviation', None)
            if d:
                dt = []
                if os.path.exists(d):
                    dt = map(float, open(d).readlines())
                else:
                    dt = map(float, d.split(','))
                GPX.dev = deviation.Deviation(dt)
            else:
                GPX.dev = None
        else:
            self.conf = parent.conf
            self.root = parent.root
            self.ns = parent.ns
            self.fields = self.conf.get('fields', gpx_fields)
        self.debug = int(self.conf.get('debug', 0))


    def __str__(self):
        return '{}: {}'.format(self.creator,self.name)

    def get_tag(self, tag):
        """return the raw element"""
        return self.node.find(self.ns + tag)

    def get_alltag(self, tag):
        """return the raw element"""
        return self.node.findall(self.ns + tag)

    def get_tag_attribute(self, tag, attrib):
        """return named attribute of tag"""
        e = self.get_tag(tag)
        if e is not None:
            try:
                return e.attrib[attrib]
            except:
                pass
        return None

    def get_tag_value(self, tag, default=None):
        """return value of tag or default"""
        e = self.get_tag(tag)
        if e is not None:
            if e.text is not None:
                return e.text.strip()
        return default

    def get_tag_set(self, tag):
        """return tag value as a set"""
        e = self.get_tag_value(tag)
        if e is not None:
            return set(e.split(' '))
        return set()

    def get_field(self, field, header):
        fname = field.lower().replace(' ','_')
        if header:
            prefix = 'header_'
        else:
            prefix = 'out_'
        try:
            m = getattr(self, prefix + fname)
            if m:
                return m()
        except:
            pass
        if header:
            return field
        try:
            m = getattr(self, fname)
            if m:
                return m
        except:
            pass
        return self.fmt_empty.format('')

    def out_fields(self,header=False):
        rc = []
        for f in self.fields.split(','):
            if f:
                rc.append(self.get_field(f, header))
                rc.append(self.sep)
        return rc

    def out_name(self):
        return self.name
    
    def __str__(self):
        return ''.join(self.out_fields())

class WPT(GPX):
    def __init__(self, wpt, parent):
        GPX.__init__(self, parent)
        self.node = wpt
        self.name = self.get_tag_value('name', 'unnamed')
        #self.time = self.get_tag_value('time')
        if self.debug:
            print('wpt name={}'.format(self.name), file=sys.stderr)
        self.latlong = latlong.LatLong(self.node.attrib['lat'],
                                       self.node.attrib['lon'],
                                       self.conf)
        self.variation = self.gm.GeoMag(self.latlong.latitude,self.latlong.longitude).dec
        self.description = self.get_tag_value('desc','').replace('\n',self.sep)
        self.magnetic = self.compass = ''

    def header_position(self):
        return self.fmt_position.format('Position')

    def out_position(self):
        return self.latlong.dm()

    def header_waypoint(self):
        return self.fmt_waypoint.format('Waypoint')

    def out_waypoint(self):
        return self.fmt_waypoint.format(self.name)

    def out_description(self):
        return self.description

class Leg(WPT):
    def __init__(self, rte, pwpt, wpt=None, leg=0):
        GPX.__init__(self, rte)
        self.name = pwpt.name
        self.leg = leg
        self.latlong = pwpt.latlong
        self.remarks = pwpt.description
        self.rte = rte
        self.variation = pwpt.variation
        if wpt:
            self.nav = latlong.Nav(pwpt.latlong, wpt.latlong, self.conf)
            self.course = self.nav.course()
            self.distance = self.nav.distance()
            self.time = self.nav.ete(rte.speed)
            self.magnetic = (self.course - self.variation) % 360.0
            if self.dev:
                self.deviation = self.dev.lookup(self.magnetic)
                self.compass = self.dev.compass(self.magnetic)
        else:
            self.nav = None

    def header_course(self, h='Course'):
        return self.hdr_course.format(h)

    def header_magnetic(self):
        return self.header_course('Magnetic')

    def header_compass(self):
        return self.header_course('Compass')

    def out_course(self):
        return self.fmt_course.format(self.course)

    def out_magnetic(self):
        return self.fmt_magnetic.format(self.magnetic)

    def out_compass(self):
        if self.dev:
            return self.fmt_compass.format(self.compass)
        return self.hdr_course(self.compass)

    def header_distance(self, h='Distance'):
        return self.hdr_distance.format(h)

    def header_cdistance(self):
        return self.header_distance('cDistance')

    def header_log(self):
        return self.header_distance('Log')

    def header_clog(self):
        return self.header_distance('cLog')

    def header_time(self, h='Time'):
        return self.hdr_time.format(h)

    def header_ete(self):
        return self.header_time('ETE')

    def out_deviation(self):
        if self.dev:
            return self.fmt_deviation.format(self.deviation)
        return self.hdr_course('')

    def out_distance(self):
        if self.nav:
            return self.fmt_distance.format(self.distance)
        return self.hdr_distance.format('')

    def out_cdistance(self):
        if self.nav and self.log_error:
            return self.fmt_distance.format(self.distance * self.log_error)
        return self.hdr_distance.format('')
    
    def out_ete(self):
        return self.fmt_time.format(self.rte.T)

    def out_leg(self):
        return self.fmt_leg.format(self.leg)

    def out_log(self):
        return self.fmt_distance.format(self.rte.L)

    def out_clog(self):
        if self.log_error:
            return self.fmt_distance.format(self.rte.L * self.log_error)
        return self.hdr_distance.format('')
    
    def out_time(self):
        if self.nav:
            return self.fmt_time.format(self.time)
        return self.hdr_time.format('')

    def out_speed(self):
        if self.nav:
            return self.fmt_speed.format(self.rte.speed)
        return self.hdr_speed.format('')

    def out_variation(self):
        if self.nav:
            return self.fmt_variation.format(self.variation)
        return self.hdr_course.format('')
        
class RTE(GPX):
    def __init__(self, rte, parent):
        GPX.__init__(self, parent)
        self.node = rte
        self.speed = self.conf.get('speed', 6.0)
        self.name = self.get_tag_value('name', 'unnamed')
        self.waypoints = []
        for rtept in self.get_alltag('rtept'):
            wpt = WPT(rtept, self)
            self.waypoints.append(wpt)

    def __str__(self):
        return self.desc()

        
    def desc(self):
        rc = []
        if self.fmt == 'verbose':
            rc.append('Route: ')
        rc.append(self.name)
        pwpt = None
        self.T = 0.0
        self.L = 0.0
        header = False
        legs = -1
        for wpt in self.waypoints:
            if pwpt:
                leg = Leg(self, pwpt, wpt, legs)
                if not header:
                    header = True
                    rc.append('\n\n')
                    rc += leg.out_fields(True)
                    rc.append('\n')
                rc.append('\n')
                rc += leg.out_fields()
                self.L += leg.distance
                self.T += leg.time
            pwpt = wpt
            legs += 1
        leg = Leg(self, pwpt, None, legs)
        rc.append('\n')
        rc += leg.out_fields()
        rc.append('\n')
        return ''.join(rc)

    def output(self):
        ofile = self.conf.get('ofile', '-')
        if ofile == '-':
            self.ofile = sys.stdout
        else:
            if ofile.endswith('-') or ofile.endswith('/'):
                ofile += '{}.{}'
            # assume first {} is for rte name and second for format
            self.ofile = open(ofile.format(self.name.replace(' ', '_'), self.fmt), 'w')
        print(self.desc(), file=self.ofile)

if __name__ == '__main__':
    import getopt
    import conf as cf

    opts,args = getopt.getopt(sys.argv[1:], 'c:D:dF:f:l:o:p:s:t:')

    conf = {
        'debug': 0,
        'fmt': 'verbose',
        'min_precision': 1,
        'ofile': '-',
        'speed': 6.0,
        }

    conf = cf.loadConfig('${progdir}/${progname:R}.rc', conf, False)
    conf = cf.loadConfig('${HOME}/.${progname:R}rc', conf, False)
    conf = cf.loadConfig('.../.${progname:R}rc', conf, False)
    
    for o,a in opts:
        if o == '-c':
            conf = cf.loadConfig(a, conf)
        elif o == '-D':
            conf['deviation'] = a
        elif o == '-d':
            conf['debug'] = int(conf.get('debug', 0)) + 1
        elif o == '-F':
            conf['fields'] = a
        elif o == '-f':
            conf['fmt'] = a
        elif o == '-l':
            conf['log_error'] = float(a)
        elif o == '-o':
            conf['ofile'] = a
        elif o == '-p':
            conf['min_precision'] = int(a)
        elif o == '-s':
            conf['speed'] = float(a)
        elif o == '-t':
            conf['sep'] = a

    for g in args:
        conf['gpx_file'] = g
        conf['gpx_f'] = open(g, 'r');
        G = GPX(conf=conf)

        print('Routes from: {}\n'.format(G))
        for rte in G.root.findall(G.ns+'rte'):
            RTE(rte, G).output()

