#!/usr/bin/env python3

from __future__ import print_function

"""Conversion of Latitude and Longitude to/from various forms

We can input and output any of

decimal degrees:

	37.36004, -122.04822

decimal minutes:

        37 21.602 N, 122 2.893 W

degrees minutes and seconds:

	37 21 36 N, 122 2 53 W

into any of the others.

The precision for decimal degrees and minutes are adjustable.

"""

"""
RCSid:
	$Id: latlong.py,v 1.20 2023/07/11 23:53:32 sjg Exp $

	@(#) Copyright (c) 2016-2023, Simon J. Gerraty

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net

"""

import sys
import math

def dm2dd(s):
    """DDD MM.mmm {N,S,E,W}"""
    w = s.strip().replace(',', '').split()
    dd = float(w[0]) + float(w[1])/60.0
    if len(w) > 2:
        ref = w[2]
        if ref in "SW":
            dd *= -1.0
    else:
        ref = ''
    return (dd,ref)

def dd2dm(dd):
    deg = int(dd)
    if deg < 0:
        dd *= -1.0
    min = (60.0 * dd) % 60.0
    return (deg,min)

class LatLong(object):
    """Store latitude and longitude in decimal degrees

    we convert input to decimal degrees and output any desired format.

    Precision of decimal degrees and minutes can be set.

    A nautical mile (1 minute of latitude) is 1852m (6076.12 feet)

    Default precision   resolution
    degrees 5 places    1.1m
    minutes 3 places    1.9m
    seconds 1 place     3.0m

    """

    def __init__(self, _lat, _long=None, conf={}):
        """save provided position as decimal degrees"""
        self.conf = conf
        self.debug = conf.get('debug', 0)
        self.lat_ref = self.long_ref = None
        self.latitude = self._latitude = 0.0
        self.longitude = self._longitude = 0.0
        self.dmout = self.dsout = None
        self.deg_prec = int(self.conf.get('deg_precision', 5))
        self.min_prec = int(self.conf.get('min_precision', 3))
        self.sec_prec = int(self.conf.get('sec_precision', 1))
        w = _lat.strip().replace(',', '').split()
        if _long:
            w += _long.strip().replace(',', '').split()
        if w[0].find('.') > 0:          # decimal degrees
            # [+-]DDD.ddddd [{N,S}][[,] [+-]DDD.ddddd [{E,W}]]
            self.latitude = self._latitude = float(w[0])
            if w[1] in ['E','W']:
                self.long_ref = self.lat_ref
                self.longitude = self._longitude = self._latitude
            elif w[1] in ['N','S']:
                self.lat_ref = w[1]
                if len(w) > 2:
                    self.long_ref = w[3]
                    self.longitude = self._longitude = float(w[2])
            elif len(w) > 1:
                self.longitude = self._longitude = float(w[1])
        elif w[1].find('.') > 0 or w[2].isalpha():        # decimal minutes
            # DDD MM[.mmm] {N,S}[[,] DDD MM[.mmm] {E,W}]
            if w[2] in ['E','W']:
                self.longitude = self._longitude = float(w[0]) + float(w[1])/60.0
                self.long_ref = w[2]
            else:
                self.latitude = self._latitude = float(w[0]) + float(w[1])/60.0
                self.lat_ref = w[2]
                if len(w) > 3:
                    self.longitude = self._longitude = float(w[3]) + float(w[4])/60.0
                    self.long_ref = w[5]
        else:                           # degrees,minutes,seconds
            # DDD MM SS[.s] {N,S}[[,] DDD MM SS[.s] {E,W}]
            if w[3] in ['E','W']:
                self.long_ref = w[3]
                self.longitude = self._longitude = float(w[0]) + float(w[1])/60.0 + float(w[2])/3600.0
            else:
                self.lat_ref = w[3]
                self.latitude = self._latitude = float(w[0]) + float(w[1])/60.0 + float(w[2])/3600.0
                if len(w) > 4:
                    self.longitude = self._longitude = float(w[4]) + float(w[5])/60.0 + float(w[6])/3600.0
                    self.long_ref = w[7]

        if not self.lat_ref:
            if self.latitude < 0.0:
                self._latitude = -1.0 * self.latitude
                self.lat_ref = 'S'
            else:
                self.lat_ref = 'N'
            if self.longitude < 0.0:
                self._longitude = -1.0 * self.longitude
                self.long_ref = 'W'
            else:
                self.long_ref = 'E'
        else:
            if self.long_ref == 'W':
                self.longitude *= -1.0
            if self.lat_ref == 'S':
                self.latitude *= -1.0

        if not self.long_ref:
            self.long_ref = 'E'

        if self.longitude < 0.0:
            # west longitude: gha is 360 increasing to west
            self.gha = abs(self.longitude)
        else:
            self.gha = 360.0 - self.longitude
            
    def dd(self):
        return '{:{width}.{prec}f}, {:{width}.{prec}f}'.format(
            self.latitude,
            self.longitude,
            width=4+self.deg_prec,
            prec=self.deg_prec)

    def dm(self):
        if not self.dmout:
            self.dmout = True
            self.lat_deg = int(self._latitude)
            self.lat_dmin = (60.0 * self._latitude) % 60.0
            self.long_deg = int(self._longitude)
            self.long_dmin = (60.0 * self._longitude) % 60.0

        return '{} {:{width}.{prec}f} {}, {} {:{width}.{prec}f} {}'.format(
            self.lat_deg,
            self.lat_dmin,
            self.lat_ref,
            self.long_deg,
            self.long_dmin,
            self.long_ref,
            width=3+self.min_prec,
            prec=self.min_prec)
                                    
    def dms(self):
        if not self.dsout:
            self.dsout = True
            self.dm()
            self.lat_min = int(self.lat_dmin)
            self.long_min = int(self.long_dmin)
            self.lat_sec = float(3600.0 * self._latitude) % 60.0
            self.long_sec = float(3600.0 * self._longitude) % 60.0

        return '{} {} {:.1f} {}, {} {} {:.1f} {}'.format(
            self.lat_deg,
            self.lat_min,
            self.lat_sec,
            self.lat_ref,
            self.long_deg,
            self.long_min,
            self.long_sec,
            self.long_ref)

    def __str__(self):
        fmt = self.conf.get('format', 'd')
        if fmt == 's':
            return self.dms()
        elif fmt == 'm':
            return self.dm()
        return self.dd()

class Nav(object):
    """Navigate from one position to another

    Given starting point A and end point B
    We can compute course distance and ETE at a given speed.

    Simple trig is not adequate.
    Formulas used are from:
    https://www.movable-type.co.uk/scripts/latlong.html
    """
    R = 6371000.0 / 1852.0         # earth radius (m converted to NMi)

    def __init__(self, a, b, conf={}):
        self.conf = conf
        self.debug = conf.get('debug', 0)
        self.a = a
        self.b = b
        self.delta_lon = b.longitude - a.longitude
        self.delta_lat = b.latitude - a.latitude
        self.h = 0
        self.c = -1
        # we need these for the math...
        self.dlat = math.radians(self.delta_lat)
        self.dlon = math.radians(self.delta_lon)
        self.lat1 = math.radians(self.a.latitude)
        self.lat2 = math.radians(self.b.latitude)

    def course(self):
        """course in degrees True"""
        if not self.c >= 0:
            y = math.sin(self.dlon) * math.cos(self.lat2)
            x = math.cos(self.lat1) * math.sin(self.lat2) - \
                math.sin(self.lat1) * math.cos(self.lat2) * \
                math.cos(self.dlon)
            self.c = math.degrees(math.atan2(y,x))
            self.c = (self.c + 360.0) % 360.0

        return self.c

    def distance(self):
        """distance in nautical miles"""
        if not self.h:
            # one degree is 60 NMi
            # spherical cosines
            self.h = math.acos(math.sin(self.lat1) * math.sin(self.lat2) +
                          math.cos(self.lat1) * math.cos(self.lat2) *
                          math.cos(self.dlon)) * self.R

        return self.h

    def ete(self, speed):
        """Estimate Time Enroute in minutes at given speed"""
        return 60.0 * self.distance() / speed
        
if __name__ == '__main__':
    import getopt
    import conf as cf

    conf = {
        'format': 'm',
        }
    conf = cf.loadConfig('${progdir}/${progname:R}.rc', conf, False)
    conf = cf.loadConfig('${HOME}/.${progname:R}rc', conf, False)
    conf = cf.loadConfig('.../.${progname:R}rc', conf, False)

    opts,args = getopt.getopt(sys.argv[1:], 'ac:dmsi:D:M:')

    inf = sys.stdin
    for o,a in opts:
        if o == '-a':
            conf['format'] = 'a'
        elif o == '-c':
            conf = cf.loadConfig(a, conf)
        elif o == '-D':
            conf['deg_precision'] = int(a)
        elif o == '-d':
            conf['format'] = 'd'
        elif o == '-i':
            inf = open(a, 'rb')
        elif o == '-M':
            conf['min_precision'] = int(a)
        elif o == '-m':
            conf['format'] = 'm'
        elif o == '-s':
            conf['format'] = 's'

    pll = None
    for line in inf:
        # make sure we can take input from google maps
        # eg. 43.3460992,-124.3262939
        ll = LatLong(line.strip().replace(',',' '), conf=conf)

        if pll:
            nav = Nav(pll, ll)
            print('course {:03}T distance {:0.2f} NMi'.format(int(nav.course()), nav.distance()))
        pll = ll
        if fmt == 'a':
            print(ll.dd())
            print(ll.dm())
            print(ll.dms())
        else:
            print(ll)

